package org.expeditee.io;

import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;

import javax.imageio.ImageIO;

import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Justification;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.reflection.JavaFX;
import org.w3c.dom.Node;
import org.w3c.dom.html.HTMLBodyElement;

/**
 * Methods to convert webpages to Expeditee frames
 * 
 * @author ngw8
 * @author jts21
 */
public class WebParser {

	
	/**
	 * Loads a webpage and renders it as Expeditee frame(s)
	 * 
	 * @param URL
	 *            Page to load
	 * @param frame
	 *            The Expeditee frame to output the converted page to
	 */
	public static void parseURL(final String URL, final Frame frame) {
		try {
			JavaFX.PlatformRunLater.invoke(null, new Runnable() {
				@Override
                public void run() {
                    try {
                    	Object webEngine = JavaFX.WebEngineConstructor.newInstance(URL);
	                    loadPage(webEngine, frame);
                    } catch (Exception e) {
	                    e.printStackTrace();
                    }
                }
			});
        } catch (Exception e) {
	        e.printStackTrace();
        }
	}
	
	protected static void loadPage(final Object webEngine, final Frame frame) throws Exception {
		JavaFX.ReadOnlyObjectPropertyAddListener.invoke(JavaFX.WorkerStateProperty.invoke(JavaFX.WebEngineGetLoadWorker
		        .invoke(webEngine)), java.lang.reflect.Proxy.newProxyInstance(
		        		JavaFX.ChangeListener.getClassLoader(), new java.lang.Class[] { JavaFX.ChangeListener },
		        new java.lang.reflect.InvocationHandler() {
        	        @Override
        	        public Object invoke(Object proxy, java.lang.reflect.Method method, Object[] args)
        	                throws java.lang.Throwable {
        		        String method_name = method.getName();
        		        // Class<?>[] classes = method.getParameterTypes();
        		        // public void changed(ObservableValue ov, State oldState, State newState)
        		        if (method_name.equals("changed")) {
        			        // changed takes 3 args
        			        if (args == null || args.length != 3) {
        				        return null;
        			        }
        			        // args[0] is the ObservableValue
        			        // args[2] is the new State
        			        if (args[2].getClass() == JavaFX.State) {
        				        int id = JavaFX.StateConstants.indexOf(args[2]);
        				        switch (id) {
        						case 0: // READY
        							// MessageBay.displayMessage("WebEngine ready");
        							break;
        						case 1: // SCHEDULED
        							// MessageBay.displayMessage("Scheduled page load");
        							break;
        						case 2: // RUNNING
        							System.out.println("Loading page!");
        							// MessageBay.displayMessage("WebEngine running");
        							break;
        						case 3: // SUCCEEDED
        							// MessageBay.displayMessage("Finished loading page");
        							System.out.println("Parsing page!");
        							JavaFX.WebEngineExecuteScript.invoke(webEngine, "window.resizeTo(800, 800)");
        							parsePage(webEngine, frame);
        							System.out.println("Parsed page!");
        							break;
        						case 4: // CANCELLED
        							MessageBay.displayMessage("Cancelled loading page");
        							break;
        						case 5: // FAILED
        							MessageBay.displayMessage("Failed to load page");
        							break;
        						}
        					}
        					System.out.println("\n");
        				}
        				return null;
        			}
                }));
	}
	
	/**
	 * Converts a loaded page to Expeditee frame(s)
	 * 
	 * @param webEngine
	 *            The JavaFX WebEngine in which the page to be converted is loaded
	 * @param frame
	 *            The Expeditee frame to output the converted page to
	 */
	public static void parsePage(final Object webEngine, final Frame frame) {
		try {
			JavaFX.PlatformRunLater.invoke(null, new Runnable() {
				@Override
				public void run() {
					try {
						HTMLBodyElement doc = (HTMLBodyElement) JavaFX.WebEngineExecuteScript.invoke(webEngine, "document.body");

						Object window = JavaFX.WebEngineExecuteScript.invoke(webEngine, "window");

						frame.setBackgroundColor(rgbStringToColor((String) JavaFX.JSObjectCall.invoke(JavaFX.JSObjectCall.invoke(window, "getComputedStyle", new Object[] { doc }), "getPropertyValue",
								new Object[] { "background-color" })));

						// Using Javascript to get an array of all the nodes in the document
						Object nodes = JavaFX.WebEngineExecuteScript.invoke(webEngine,
								"function getTextNodes(rootNode){" 
									+ "var node;" + "var textNodes=[];"
									+ "var walk = document.createTreeWalker(rootNode, NodeFilter.SHOW_ALL);"
									+ "while(node=walk.nextNode()) {" 
										+ " textNodes.push(node);" + "}" 
									+ "return textNodes;"
									+ "}; " 
								+ "getTextNodes(document.body)");

						int nodesLength = (Integer) JavaFX.JSObjectGetMember.invoke(nodes, "length");

						for (int i = 0; i < nodesLength; i++) {
							Node currentNode = (Node) JavaFX.JSObjectGetSlot.invoke(nodes, i);
							
							if (currentNode.getNodeType() == Node.TEXT_NODE || currentNode.getNodeType() == Node.ELEMENT_NODE) {

								System.out.println(i + "/" + nodesLength + " : " + currentNode);
								
								Object style;
								Object bounds;

								if (currentNode.getNodeType() == Node.TEXT_NODE) {
									// CSS style for the element
									style = JavaFX.JSObjectCall.invoke(window, "getComputedStyle", new Object[] { currentNode.getParentNode() });

									// Getting a rectangle that represents the area and position of the element
									bounds = JavaFX.JSObjectCall.invoke(currentNode.getParentNode(), "getBoundingClientRect", new Object[] {});
								} else {
									style = JavaFX.JSObjectCall.invoke(window, "getComputedStyle", new Object[] { currentNode });

									bounds = JavaFX.JSObjectCall.invoke(currentNode, "getBoundingClientRect", new Object[] {});
								}
								
								// Bounding rectangle position is relative to the current view, so scroll position must be added to x/y
								float x = Float.valueOf(JavaFX.JSObjectGetMember.invoke(bounds, "left").toString())
										+ Float.valueOf(JavaFX.WebEngineExecuteScript.invoke(webEngine, "window.pageXOffset").toString());
								float y = Float.valueOf(JavaFX.JSObjectGetMember.invoke(bounds, "top").toString())
										+ Float.valueOf(JavaFX.WebEngineExecuteScript.invoke(webEngine, "window.pageYOffset").toString());

								float width = Float.valueOf(JavaFX.JSObjectGetMember.invoke(bounds, "width").toString());
								float height = Float.valueOf(JavaFX.JSObjectGetMember.invoke(bounds, "height").toString());

								// Checking if the element is actually visible on the page
								if (WebParser.elementVisible(x, y, width, height, style)) {
	
									// Filtering the node type, starting with text nodes
									if (currentNode.getNodeType() == Node.TEXT_NODE) {
										String fontSize = ((String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "font-size" }));

										// Trimming off the units (always px) from the font size
										fontSize = fontSize.substring(0, fontSize.length() - 2);
										
										// Always returns in format "rgb(x,x,x)" or "rgba(x,x,x,x)"
										String color = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "color" });
										
										// Always returns in format "rgb(x,x,x)" or "rgba(x,x,x,x)"
										String bgColor = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "background-color" });
										
										String align = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "text-align" });
										
										// Returns comma-separated list of typefaces
										String typeface = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "font-family" });
										String[] typefaces = typeface.split(", |,");
										
										String weight = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "font-weight" });
										
										String fontStyle = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "font-style" });

										// Returns "normal" or a value in pixels (e.g. "10px")
										String letterSpacing = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "letter-spacing" });

										Boolean fontFound = false;
										Font font = new Font(null);

										// Looping through all font-families listed in the element's CSS until one that is installed is
										// found, or the end of the list is reached, in which case the default font is used
										for (int j = 0; j < typefaces.length && !fontFound; j++) {

											// Regex will remove any inverted commas surrounding multi-word typeface names
											font = new Font(typefaces[j].replaceAll("^'|'$", ""), Font.PLAIN, 12);

											if (!(font.getFamily().toLowerCase().equals(Font.DIALOG.toLowerCase()))) {
												fontFound = true;
											}
										}

										String fontStyleComplete = "";

										int weightInt = 0;

										try {
											weightInt = Integer.parseInt(weight);
										} catch (NumberFormatException nfe) {
											weightInt = 0;
										}

										// checking if font is bold - i.e. 'bold', 'bolder' or weight over 500
										if (weight.toLowerCase().startsWith("bold") || weightInt > 500) {
											fontStyleComplete = fontStyleComplete.concat("bold");
										}

										if (fontStyle.toLowerCase().equals("italic") || fontStyle.toLowerCase().equals("oblique")) {
											fontStyleComplete = fontStyleComplete.concat("italic");
										}

										int letterSpacingInt = 0;

										try {
											letterSpacingInt = (int) (Integer.parseInt(letterSpacing.substring(0, letterSpacing.length() - 2)) / Float.valueOf(fontSize));
										} catch (NumberFormatException nfe) {
											letterSpacingInt = 0;
										}

										Text t;

										t = frame.addText((int) x, (int) y, currentNode.getTextContent().replaceAll("\\s+", " ").trim(), null);

										t.setColor(rgbStringToColor(color));
										t.setBackgroundColor(rgbStringToColor(bgColor));
										t.setFont(font);
										t.setSize(Float.valueOf(fontSize));
										t.setFontStyle(fontStyleComplete);
										t.setLetterSpacing(letterSpacingInt);

										if (align.equals("left")) {
											t.setJustification(Justification.left);
										} else if (align.equals("right")) {
											t.setJustification(Justification.right);
										} else if (align.equals("center")) {
											t.setJustification(Justification.center);
										} else if (align.equals("justify")) {
											t.setJustification(Justification.full);
										}

										// Font size divided by 2 is added to the item width to give a little breathing room
										t.setWidth(Math.round(width + (t.getSize() / 2)));

									} else if (currentNode.getNodeType() == Node.ELEMENT_NODE) {

										// background image, returns in format "url(protocol://absolute/path/to/img.extension)" for images,
										// may also return gradients, data, etc. (not handled yet). Only need to add bg image on
										// 'ELEMENT_NODE' (and not 'TEXT_NODE' otherwise there would be double-ups
										String bgImage = (String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "background-image" });

										if (bgImage.startsWith("url(")) {

											bgImage = bgImage.substring(4, bgImage.length() - 1);

											String bgSize = ((String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "background-size" })).toLowerCase();

											// Returns "[x]px [y]px", "[x]% [y]%", "[x]px [y]%" or "[x]% [y]px"
											String bgPosition = ((String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "background-position" })).toLowerCase();

											String[] bgOffsetCoords = bgPosition.split(" ");

											int bgOffsetX = 0, bgOffsetY = 0;

											int cropStartX, cropStartY, cropEndX, cropEndY;

											// Converting the x and y offset values to integers (and from % to px if needed)
											if (bgOffsetCoords[0].endsWith("%")) {
												bgOffsetX = (int) ((Integer.valueOf(bgOffsetCoords[0].substring(0, bgOffsetCoords[0].length() - 1)) / 100.0) * width);
											} else if (bgOffsetCoords[0].endsWith("px")) {
												bgOffsetX = (int) (Integer.valueOf(bgOffsetCoords[0].substring(0, bgOffsetCoords[0].length() - 2)));
											}

											if (bgOffsetCoords[1].endsWith("%")) {
												bgOffsetY = (int) ((Integer.valueOf(bgOffsetCoords[1].substring(0, bgOffsetCoords[1].length() - 1)) / 100.0) * height);
											} else if (bgOffsetCoords[1].endsWith("px")) {
												bgOffsetY = (int) (Integer.valueOf(bgOffsetCoords[1].substring(0, bgOffsetCoords[1].length() - 2)));
											}

											// Converting from an offset to crop coords
											cropStartX = -1 * bgOffsetX;
											cropEndX = (int) (cropStartX + width);

											cropStartY = -1 * bgOffsetY;
											cropEndY = (int) (cropStartY + height);

											int bgWidth = -1;

											if (bgSize.equals("cover")) {
												bgWidth = (int) width;
											} else if (bgSize.equals("contain")) {
												// TODO: actually compute the appropriate width
												bgWidth = (int) width;
											} else if (bgSize.equals("auto")) {
												bgWidth = -1;
											} else {
												bgSize = bgSize.split(" ")[0];

												if (bgSize.endsWith("%")) {
													bgWidth = (int) ((Integer.parseInt(bgSize.replaceAll("\\D", "")) / 100.0) * width);
												} else if (bgSize.endsWith("px")) {
													bgWidth = Integer.parseInt(bgSize.replaceAll("\\D", ""));
												}
											}

											try {
												WebParser.addImageFromUrl(bgImage, frame, x, y, bgWidth, cropStartX, cropStartY, cropEndX, cropEndY);
											} catch (MalformedURLException mue) {
												// probably a 'data:' url, not supported yet
												mue.printStackTrace();
											}
										}

										String imgSrc;

										if (currentNode.getNodeName().toLowerCase().equals("img") && (imgSrc = JavaFX.JSObjectGetMember.invoke(currentNode, "src").toString()) != null) {
											try {
												WebParser.addImageFromUrl(imgSrc, frame, x, y, (int) width, null, null, null, null);
											} catch (MalformedURLException mue) {
												// probably a 'data:' url, not supported yet
												mue.printStackTrace();
											}
										}
									}
								}
							}
						}

					} catch (Exception e) {
						e.printStackTrace();
					}
					System.out.println("Parsed frame");
					FrameIO.SaveFrame(frame);
				}
			});
		} catch (Exception e) {
			e.printStackTrace();
		}
    }

	/**
	 * @param rgbString
	 *            string in the format <i>rgb(x,x,x)</i> or <i>rgba(x,x,x,x)</i>
	 * @return A Color object that should match the rgb string passed int. Returns null if alpha is 0
	 */
	private static Color rgbStringToColor(String rgbString) {
		// Splitting the string into 'rgb' and 'x, x, x'
		String[] tmpStrings = rgbString.split("\\(|\\)");

		// Splitting up the RGB(A) components into an array
		tmpStrings = tmpStrings[1].split(",");

		int[] components = new int[4];
		Arrays.fill(components, 255);

		for (int i = 0; i < tmpStrings.length; i++) {
			Float d = Float.parseFloat(tmpStrings[i].trim());

			components[i] = Math.round(d);
		}

		if (components[3] > 0) {
			return new Color(components[0], components[1], components[2], components[3]);
		} else {
			return null;
		}
	}

	private static boolean elementVisible(float x, float y, float width, float height, Object style) {
		try {
			if (width <= 0 || height <= 0 || x + width <= 0 || y + height <= 0 || ((String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "visibility" })).equals("hidden")
					|| ((String) JavaFX.JSObjectCall.invoke(style, "getPropertyValue", new Object[] { "display" })).equals("none")) {
				return false;
			} else {
				return true;
			}
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * @param imgSrc
	 *            URL of the image to add
	 * @param frame
	 *            Frame to add the image to
	 * @param x
	 *            X-coordinate at which the image should be placed on the frame
	 * @param y
	 *            Y-coordinate at which the image should be placed on the frame
	 * @param width
	 *            Width of the image once added to the frame. Negative 1 (-1) will cause the actual width of the image file to be used
	 * 
	 * @param cropStartX
	 *            X-coordinate at which to start crop, or null for no crop
	 * @param cropStartY
	 *            Y-coordinate at which to start crop, or null for no crop
	 * @param cropEndX
	 *            X-coordinate at which to end the crop, or null for no crop
	 * @param cropEndY
	 *            Y-coordinate at which to end the crop, or null for no crop
	 * @throws MalformedURLException
	 * @throws IOException
	 */
	private static void addImageFromUrl(String imgSrc, final Frame frame, float x, float y, int width, Integer cropStartX, Integer cropStartY, Integer cropEndX, Integer cropEndY)
			throws MalformedURLException,
			IOException {

		URL imgUrl = new URL(imgSrc);

		HttpURLConnection connection = (HttpURLConnection) (imgUrl.openConnection());

		// Spoofing a widely accepted User Agent, since some sites refuse to serve non-webbrowser clients
		connection.setRequestProperty("User-Agent", "Mozilla/5.0");

		BufferedImage img = ImageIO.read(connection.getInputStream());

		int hashcode = Arrays.hashCode(img.getData().getPixels(0, 0, img.getWidth(), img.getHeight(), (int[]) null));
		File out = new File(FrameIO.IMAGES_PATH + Integer.toHexString(hashcode) + ".png");
		out.mkdirs();
		ImageIO.write(img, "png", out);

		if (cropEndX == null || cropStartX == null || cropEndY == null || cropStartY == null) {
			cropStartX = 0;
			cropStartY = 0;
			cropEndX = img.getWidth();
			cropEndY = img.getHeight();
		}

		if (width < 0) {
			width = img.getWidth();
		}

		Picture pic = new Picture(out.getName(), frame);
		pic.setWidth(width);

		// Have to divide the crop coords by the image scale, since Expeditee seems to always crop then scale
		pic.setCropStart(new Point((int) (cropStartX / pic.getScale()), (int) (cropStartY / pic.getScale())));
		pic.setCropEnd(new Point((int) (cropEndX / pic.getScale()), (int) (cropEndY / pic.getScale())));
		pic.setPosition(x, y);
		frame.addItem(pic);
	}
}
